#packages
using DataFrames
using DelimitedFiles # allows reading and writing text files with delimiters
using XLSX
using JuMP          #
using Gurobi        # Gurobi solver (requires a functional Gurobi installation)

#A. Reading the data required
#A.1. Classrooms
classrooms_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "Classrooms"))

#A.2. Programs
programs_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "Programs"))

#A.3. Schedules
schedules_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "Schedules"))

#A.4. Schedules
general_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "General"))

#A.5. Availability
availability_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "Availability"))

#A.6. Type
type_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "Type"))

#A.7. Students
students_table = DataFrame(XLSX.readtable("Thesis/Information_needed - 3.C.xlsx", "Students"))


#Definition of the indices:
R = classrooms_table[1:21,2] #classrooms
D = schedules_table[1:5,2] #days
T = schedules_table[1:24,1] #timeslots
I = general_table[1:22,1] #instructors
S = students_table[1:4,1] #students' groups
C = programs_table[1:42,1] #courses

# Get the number of corresponding indices
nb_rooms = size(classrooms_table, 1)
nb_days = 5 
nb_timeslots = size(schedules_table,1)
nb_instructors = size(general_table,1)
nb_students = size(students_table,1) 
nb_courses = size(programs_table,1)

# Parameters
# STUDENT_GROUP_SIZE
# Create a dictionary to map course names to indices

course_numbers = Dict(programs_table.Courses[i] => i for i in 1:nb_courses)

# Define student_group_size parameter
student_group_size = Dict(course_numbers[programs_table.Courses[i]] => programs_table.number_of_students[i] for i in 1:nb_courses)

# Now we can access student group sizes using course indices
println(student_group_size[1])

#ROOM_CAPACITY
room_mapping = Dict(row.Classroom => row.Capacity for row in eachrow(classrooms_table)) #give the corresponding room capacity
# Create a dictionary to map course names to indices
r_numbers = Dict(classrooms_table[1:nb_rooms, 2] .=> 1:nb_rooms) #replace classroom name by indices
# Create the parameter room_capacity[r]
room_capacity = Dict(r_numbers[r] => room_mapping[r] for r in keys(r_numbers)) #give the capacity for each index
#Test
println(room_capacity[1])

# DURATION(C)
duration_mapping = Dict(row.Courses => row.Duration_in_timeslots for row in eachrow(programs_table)) #associate the duration to each course
# Create a dictionary to map course names to indices
c_numbers = Dict(programs_table[1:nb_courses, 1] .=> 1:nb_courses) #replace each course name with an index
# Create the parameter duration[c]
duration = Dict(c_numbers[c] => duration_mapping[c] for c in keys(c_numbers)) #give the corresponding duration of each index
#Test
println(duration[6])

# ASSOC(s, c)
# Extract course names from programs_table
courses = programs_table[:, 1]
# Initialize Assoc with zeros
Assoc = zeros(Int, nb_students, nb_courses)

# Update Assoc based on the data in students_table
for row in 1:size(students_table, 1)
    student = students_table[row, 1]
    for col in 2:size(students_table, 2)
        course = courses[col - 1]  # Assuming course names start from the second column
        assoc_value = students_table[row, col]  # Assuming association values are 1 or 0
        
        # Find indices of the student and course
        student_index = findfirst(x -> x == student, S)
        course_index = findfirst(x -> x == course, C)
        
        # If both student and course are found, update Assoc
        if student_index ≠ nothing && course_index ≠ nothing
            Assoc[student_index, course_index] = assoc_value
        end
    end
end

#Test
println(Assoc[1, 2]) 


# COMP(i, c)
# Get the unique instructors and courses
instructors = unique(programs_table.Instructors) #to not take may times the same instructor
courses = unique(programs_table.Courses) # to not take many times the same course
# Create dictionaries to map instructor and course names to indices
i_numbers = Dict(instructors .=> 1:length(instructors)) #replace each name by an index
c_numbers = Dict(courses .=> 1:length(courses)) #replace each course name by an index
# Create the binary parameter Comp(i, c)
Comp = Dict((i_numbers[i], c_numbers[c]) => any((strip.(programs_table.Instructors) .== strip(i)) .& (strip.(programs_table.Courses) .== strip(c))) for i in instructors, c in courses)
#Test
example_result = Comp[(i_numbers["Valérie Henry"], c_numbers["Mathématiques: Analyse infinitésimale - 1"])]
println("Example Result: ", example_result)
println(Comp[(1, 1)])

#A(i,t,d): binary parameter that takes value 1 if an instructor i is available at timeslot t of day d.
# Initialize A with zeros
A = zeros(Int, nb_instructors, nb_timeslots, nb_days)

for line in 1:size(availability_table, 1)
    for i in 1:nb_instructors
        j = findfirst(x -> x == availability_table[line, :timeslots], T)
        k = findfirst(x -> x == availability_table[line, :days], D)
        availability_value = availability_table[line, i + 2]  
        # Check if indices are valid
        if !isnothing(j) && !isnothing(k)
            # Update A based on availability value
            A[i, j, k] += availability_value
        else
            println("Invalid indices for line $line: $j, $k")
        end
    end
end

A
#Test
println(A[2,2,1])

#Type(c,c’): binary parameter that takes value 1 if a course c is linked to a course c’
# Initialize Type(c, c') with zeros
type = zeros(Int, nb_courses, nb_courses)

# Set the values based on the Type table
for row in 1:size(type_table, 1)
    course1 = type_table[row, 1]
    course2 = type_table[row, 2]
    value = type_table[row, 3]
    
    # Find indices of the courses
    index1 = findfirst(x -> x == course1, C)
    index2 = findfirst(x -> x == course2, C)
    
    # If both courses are found, set the binary parameter value
    if index1 ≠ nothing && index2 ≠ nothing
        type[index1, index2] = value
        
        # Check if the value is 1, and set the symmetric pair if true
        if value == 1
            type[index2, index1] = 1
        end
    end
end
#Test
type[8,9]
type[9,8]


# ILP Gurobi Model
#CHANGE THIS TO 5 MINUTES LATER
ILP = Model(optimizer_with_attributes(Gurobi.Optimizer,"TimeLimit"=>1800)) 

#definition of the variables
@variable(ILP,X[c=1:nb_courses,i=1:nb_instructors,t=1:nb_timeslots,d=1:nb_days,r=1:nb_rooms,s=1:nb_students],Bin) #the course is given
@variable(ILP,Y[c=1:nb_courses,t=1:nb_timeslots,d=1:nb_days],Bin) #the course starts
@variable(ILP,Z[c=1:nb_courses,r=1:nb_rooms],Bin) #the rooms is associated to the course
@variable(ILP,W[i=1:nb_instructors,d=1:nb_days],Bin) #the instructor i works on day d
@variable(ILP, V[t=1:nb_timeslots, d=1:nb_days, s=1:nb_students], Int, lower_bound=0) #course conflicts


#Objective function
@objective(ILP,Min,sum(V[t,d,s] for t in 1:nb_timeslots, d in 1:nb_days, s in 1:nb_students))


# C1: ensure a course is given in one block of time, starting at the first time slot
for c in 1:nb_courses
    for d in 1:nb_days
        for t in 1:nb_timeslots
            for tt in t:(min((t+duration[c]-1),nb_timeslots))
                for s in 1:nb_students
                    @constraint(ILP, Y[c, t, d]*Assoc[s,c] <= sum(X[c, i, tt, d, r, s] for i in 1:nb_instructors, r in 1:nb_rooms))
                end
            end
        end
    end
end  



# C2: ensure the correspondance between instructor and course
for c in 1:nb_courses
    for i in 1:nb_instructors
        for s in 1:nb_students
            @constraint(ILP, sum(X[c, i, t, d, r, s] for d in 1:nb_days, r in 1:nb_rooms, t in 1:nb_timeslots) <= Comp[i, c]*duration[c])
        end
    end
end

#C3: ensure the correspondance between the student and the course
for c in 1:nb_courses
    for s in 1:nb_students
        @constraint(ILP, sum(X[c,i,t,d,r,s] for d in 1:nb_days, r in 1:nb_rooms, i in 1:nb_instructors, t in 1:nb_timeslots) == Assoc[s,c]*duration[c])
    end
end

# C4: ensure that a course starts exactly one
#OK
for c in 1:nb_courses
    @constraint(ILP, sum(Y[c, t, d] for t in 1:nb_timeslots, d in 1:nb_days) == 1)
end

#C5: ensure the sufficient capacity for a course
for r in 1:nb_rooms
    for t in 1:nb_timeslots
        for d in 1:nb_days
            for c in 1:nb_courses
                @constraint(ILP, sum(X[c,i,t,d,r,s] for i in 1:nb_instructors, s in 1:nb_students)*student_group_size[c] <= room_capacity[r])
            end
        end
    end
end

#C6: ensure that an instructor can give only one course at a time
for i in 1:nb_instructors
    for t in 1:nb_timeslots
        for d in 1:nb_days
            @constraint(ILP, sum(X[c,i,t,d,r,s] for c in 1:nb_courses, r in 1:nb_rooms, s in 1:nb_students)<=sum(Assoc[s,c] for c in 1:nb_courses, s in 1:nb_students))
        end
    end
end

#C7: CANCELLED ensure that a students' group can only be given one course at a time

#C8: Y(c,t,d)=0 
for c in 1:nb_courses
    for d in 1:nb_days
        @constraint(ILP,sum(Y[c,t,d] for t in (nb_timeslots-duration[c]+1):nb_timeslots) == 0)
    end
end

# C9: ensure that the classroom for the course is the same for the whole duration
for c in 1:nb_courses
    for r in 1:nb_rooms
        @constraint(ILP, sum(X[c, i, t, d, r, s] for i in 1:nb_instructors, t in 1:nb_timeslots, d in 1:nb_days, s in 1:nb_students) == duration[c] * sum(Assoc[s, c] for s in 1:nb_students) * Z[c, r])
    end
end

#C10 : ensure that only one classroom can be associated to a course
for c in 1:nb_courses
    @constraint(ILP, sum(Z[c, r] for r in 1:nb_rooms) == 1)
end
    
    

#NEW CONSTRAINTS of the second version
#C11: ensure that courses cannot be given from 8 to 9, from 12 to 13 and from 18 to 20

for t in vcat(1:2, 9:10, 21:24)
    for c in 1:nb_courses
        for i in 1:nb_instructors
            for d in 1:nb_days
                for r in 1:nb_rooms
                    for s in 1:nb_students
                        @constraint(ILP, sum(X[c,i,t,d,r,s]) ==0)
                    end
                end
            end
        end
    end
end

#C12: ensure that an instructor I can only get a course assigned at timeslot t of day d if they are available at that moment
for i in 1:nb_instructors
    for t in 1:nb_timeslots
        for d in 1:nb_days
            for s in 1:nb_students
                @constraint(ILP, sum(X[c,i,t,d,r,s] for c in 1:nb_courses, r in 1:nb_rooms)<=A[i,t,d])
            end
        end
    end
end

#C13: ensures that if instructor i gives at least one course on day d, then, this instructor will be considered as working on that day
for i in 1:nb_instructors
    for d in 1:nb_days
        @constraint(ILP, (sum(Y[c,t,d] for t in 1:nb_timeslots, c in 1:nb_courses)*(Comp[i,c]))<=((W[i,d])*nb_timeslots))
    end
end

#C14: ensures that an instructor cannot work more than 4 dats per week
for i in 1:nb_instructors
    @constraint(ILP, sum(W[i,d] for d in 1:nb_days)<=4)
end

#C15: ensures that if 2 courses are linked, then, the courses cannot be given within the same day
for d in 1:nb_days
    for c in 1:nb_courses
        for co in 1:nb_courses
            @constraint(ILP, (sum(Y[c,t,d] for t in 1:nb_timeslots)*type[c,co] + sum(Y[co,t,d] for t in 1:nb_timeslots)*type[c,co])<=1)
        end
    end
end

#NEW CONSTRAINTS - Version 3
#C16: scheduling conflicts
for t in 1:nb_timeslots
    for d in 1:nb_days
        for s in 1:nb_students
            @constraint(ILP, sum(X[c,i,t,d,r,s] for c in 1:nb_courses, i in 1:nb_instructors, r in 1:nb_rooms) <= V[t,d,s] + 1)
        end
    end
end

#print(ILP)
status = optimize!(ILP)
#à adapter pour mon problème
solX = value.(X)
solY = value.(Y)
solZ=value.(Z)
solW=value.(W)
solV=value.(V)

println(status)
for c in 1:nb_courses
    for d in 1:nb_days
        for t in 1:nb_timeslots
            if solY[c, t, d] ≈ 1.0
                println("Y[$c, $t, $d] = 1.0")
            end
        end
    end
end

for c in 1:nb_courses
    for d in 1:nb_days
        for t in 1:nb_timeslots
            for i in 1:nb_instructors
                for r in 1:nb_rooms
                    for s in 1:nb_students
                        if solX[c, i, t, d, r, s] ≈ 1.0
                            println("X[$c, $i, $t, $d, $r, $s] = 1.0")
                        end
                    end
                end
            end
        end
    end
end

for c in 1:nb_courses
    for r in 1:nb_rooms
        if solZ[c,r]  ≈ 1.0
            println("Z[$c, $r] = 1.0")
        end
    end
end

for i in 1:nb_instructors
    for d in 1:nb_days
        if solW[i, d] ≈ 1.0
            println("W[$i, $d] = 1.0")
        end
    end
end

for t in 1:nb_timeslots
    for d in 1:nb_days
        for s in 1:nb_students
            if solV[t,d,s]≈ 1.0
                println("V[$t, $d, $s] = 1.0")
            end
        end
    end
end
